'''
Created on 7. jan. 2016 / 19. feb. 2016 / 2. mar. 2016

Computing temporal betweenness

@author: Selena Praprotnik
'''

import math
import time
from TQclean import *

# test temporal network (Figure 4)
A = [ [ [] for i in range(5)] for j in range(5)]
A[0][1] = [(1,2,3),(2,4,2),(4,5,3),(5,7,2)]
A[1][2] = [(1,2,5),(2,3,3),(3,5,4),(5,7,2)]
A[0][4] = [(1,2,4),(2,4,3),(4,7,2)]
A[4][1] = [(1,2,1),(2,4,3),(4,7,2)]
A[4][3] = [(1,2,2),(2,4,1),(4,7,2)]
A[2][3] = [(1,2,1),(2,4,2),(4,7,1)]
A[1][3] = [(1,2,2),(2,7,1)]

A[1][0] = [(1,2,3),(2,4,2),(4,5,3),(5,7,2)]
A[2][1] = [(1,2,5),(2,3,3),(3,5,4),(5,7,2)]
A[4][0] = [(1,2,4),(2,4,3),(4,7,2)]
A[1][4] = [(1,2,1),(2,4,3),(4,7,2)]
A[3][4] = [(1,2,2),(2,4,1),(4,7,2)]
A[3][2] = [(1,2,1),(2,4,2),(4,7,1)]
A[3][1] = [(1,2,2),(2,7,1)]


# transform latency to first arrival times

# all possible arrivals, a temporal quantity; returns the table of arrivals - c[i] => arrival if you start at the time i+1
# add, check whether the arrival time is possible
def arrival(a):
    (tMin,tMax,vMin,vMax) = TQ.TqSummary(a)
    a = TQ.fillGaps(a,tMin,tMax,TQ.inf)
    c = []
    for (s,f,v) in a:
        t = s
        while t < f:
            c.append(t+v)
            t = t+1
    return c

# compute first arrival temporal quantity
def firstArrival(a):
    arr = arrival(a)
    c = []
    for i in range(len(arr)):
        c.append((i+1,i+2,min(arr[i:])))
    return TQ.standard(c)

# compute the first arrival and count the number of possible first arrivals;
# temporal quantities, values are pairs ( first arrival, number of different first arrivals )
def countArrival(a):
    arr = arrival(a)
    c = []
    for i in range(len(arr)):
        m = min(arr[i:])
        c.append((i+1,i+2,(m,arr[i:].count(m))))
    return TQ.standard(c)
    

# first arrival semiring operations
# sequential links - first traverse a, then b
# a and b are temporal quantities that tell the first arrival times
def fAprod(a,b):
    c = []
    for (sa,fa,va) in a:
        for (sb,fb,vb) in b:
            if sb <= va < fb:
                c.append((sa,fa,vb))
                break
    return TQ.standard(c)

# parallel links
# a and b are temporal quantities that tell the first arrival times
def fAsum(a,b):
      if len(a) == 0: return(b)
      if len(b) == 0: return(a)
      c = []; A = a.__iter__(); B = b.__iter__()
      (sa,fa,va) = TQ.get(A); (sb,fb,vb) = TQ.get(B)
      while (sa<TQ.inf) or (sb<TQ.inf):
         if sa < sb:
            sc = sa; vc = va
            if sb < fa: fc = sb; sa = sb
            else: fc = fa; (sa,fa,va) = TQ.get(A)
         elif sa == sb:
            sc = sa; fc = min(fa,fb); vc = min(va,vb)
            sa = sb = fc; fA = fa
            if fA <= fb: (sa,fa,va) = TQ.get(A)
            if fb <= fA: (sb,fb,vb) = TQ.get(B)
         else:
            sc = sb; vc = vb
            if sa < fb: fc = sa; sb = sa
            else: fc = fb; (sb,fb,vb) = TQ.get(B)
         c.append((sc,fc,vc))
      return(TQ.standard(c))


#traveling combinatorial semiring
#parallel links
# a and b are temporal quantities that tell the (first arrival, number of first arrivals)
# we get them from latency using the function countArrival
def tCsum(a,b):
    if len(a) == 0: return b
    if len(b) == 0: return a
    
    c = []; A = a.__iter__(); B = b.__iter__()
    (sa,fa,va) = TQ.get(A); (sb,fb,vb) = TQ.get(B)
    while (sa<TQ.inf) or (sb<TQ.inf):
        if sa < sb:
            sc = sa; vc = va
            if sb < fa: fc = sb; sa = sb
            else: fc = fa; (sa,fa,va) = TQ.get(A)
        elif sa == sb:
            sc = sa; fc = min(fa,fb); vc = TQ.geoAdd(va,vb)
            sa = sb = fc; fA = fa
            if fA <= fb: (sa,fa,va) = TQ.get(A)
            if fb <= fA: (sb,fb,vb) = TQ.get(B)
        else:
            sc = sb; vc = vb
            if sa < fb: fc = sa; sb = sa
            else: fc = fb; (sb,fb,vb) = TQ.get(B)
        c.append((sc,fc,vc))
    return(TQ.standard(c))

# sequential links
# a and b are temporal quantities that tell the (first arrival, number of first arrivals)
# we get them from latency using the function countArrival
def tCprod(a,b):    
    c = []
    for (sa,fa,va) in a:
        for (sb,fb,vb) in b:
            (fva,cva) = va; (fvb,cvb) = vb;
            if sb <= fva < fb:
                c.append((sa,fa,(fvb, cva*cvb)))
                break
    return TQ.standard(c)

# strict closure over the traveling semiring;
# the elements of the matrix tell the (first arrival time, number of first arrivals)
# we get them from latency using the function countArrival
def tempClosure(R):
    nr = len(R); nc = len(R[0])
    if nr != nc: raise TQ.TQerror("MatClosure: square matrix required")
    C = deepcopy(R)    
    for k in range(nr):
        for u in range(nr):
            for v in range(nr):
                C[u][v] = tCsum(C[u][v],tCprod(C[u][k],C[k][v]))
    return(C)


# first arrival betweenness, the elements of the temporal matrix A are latencies
def tempBetween(A):
    # transform the latencies to (first arrival, number of first arrivals)
    nr = len(A); nc = len(A[0])
    time = TQ.MatSummary(A)[1]
    B = deepcopy(A)    
    for u in range(nr):
        for v in range(nc):
            B[u][v]= countArrival(TQ.fillGaps(A[u][v],1,time))
    C = tempClosure(B)
    bw = [ [] for i in range(nr) ]
    fac = [(1,TQ.inf,1/(nr-1)/(nr-2))]
    for v in range(nr):
        b = []
        for u in range(nr):
            for w in range(nr):
                if (C[u][w]!=[]) and (u != w) and (u != v) and (v != w):
                    b = TQ.sum(b,betweenT(C[u][v],C[v][w],C[u][w]))
        bw[v] = TQ.prod(b,fac)
    return bw

def betweenT(uv,vw,uw):
    if len(uv) == 0: return ([])
    if len(vw) == 0: return ([])
    if len(uw) == 0: return ([])
    r = []
    allr = tCprod(uv,vw)
    A = uw.__iter__(); B = allr.__iter__()
    (sa,fa,va) = TQ.get(A); (sb,fb,vb) = TQ.get(B)
    while (sa<TQ.inf) or (sb<TQ.inf):
        if fa <= sb: (sa,fa,va) = TQ.get(A)
        elif fb <= sa: (sb,fb,vb) = TQ.get(B)
        else:
            sr = max(sa,sb); fr = min(fa,fb)    
            if (va[0] == vb[0]): # if the first arrival times are equal
                vr = vb[1] # we remember the result -- this is the value of n_{uw(v)}(t)
                r.append((sr,fr,vr/va[1])) # divided with n_{uw}(t)
            if fr == fa: (sa,fa,va) = TQ.get(A)
            if fr == fb: (sb,fb,vb) = TQ.get(B)
    return(TQ.standard(r))
    
def deleteInf(a):
    c = []
    for (s,f,v) in a:
        if v[0] < TQ.inf: c.append((s,f,v))
    return c

# A matrix of latencies, ind tells which node to take out
def closure2(A,ind):
    time = TQ.MatSummary(A)[1]
    # transform the latencies to (first arrival, number of first arrivals)
    nr = len(A); nc = len(A[0])
    B = deepcopy(A)    
    for u in range(nr):
        for v in range(nc):
            if (u == ind or v == ind): B[u][v] = []
            else: B[u][v]= countArrival(TQ.fillGaps(A[u][v],1,time))
    C = tempClosure(B)
    return C

def countPaths(b):
    fJ = 0
    v0 = TQ.inf
    for (s,f,v) in b:
        if v0 != v[0]: fJ = fJ + v[1] # the if is necessary if one of the first journeys starts at 1, the other at 2, they are both counted at the time 1, and one is also counted at 1,...
        v0 = v[0]
    return fJ

# first arrival betweenness with exclusion
def betweenTimeEx(A):
    # transform the latencies to (first arrival, number of first arrivals)
    nr = len(A); nc = len(A[0])
    time = TQ.MatSummary(A)[1]
    B = deepcopy(A)    
    for u in range(nr):
        for v in range(nc):
            B[u][v]= countArrival(TQ.fillGaps(A[u][v],1,time)) #change the time to relevant numbers
    C = tempClosure(B)
    
    bw = [ [] for i in range(nr) ]
    for v in range(nr):
        b = 0
        D = closure2(A,v)
        for u in range(nr):
            for w in range(nr):
                if (D[u][w]!=[]) and (u != w) and (u != v) and (v != w):
                    uv = deleteInf(C[u][v])
                    vw = deleteInf(C[v][w])
                    uw = deleteInf(D[u][w])
                    allr = deleteInf(tCprod(uv,vw))
                    uvw = countPaths(allr)
                    nuw = countPaths(uw)
                    if (uvw + nuw != 0): b = b + uvw/(uvw+nuw)
        bw[v] = b/(nr-1)/(nr-2)
    return bw



print('article ImaMat results \n')
a = [(1,2,3),(2,4,2),(4,5,3),(5,7,2)]
b = [(1,2,5),(2,3,3),(3,5,4),(5,7,2)]
c = [(1,2,4),(2,4,3),(4,7,2)]
d = [(1,2,1),(2,4,3),(4,7,2)]
print('arrival')
print(arrival(a))
print(arrival(b))

print('firstArrival')
print(firstArrival(b))

print('first arrival semiring operations')
print(fAsum(firstArrival(a),fAprod(firstArrival(c),firstArrival(d))))

print('countArrival')  
print(countArrival(a))
print(countArrival(b))

print('traveling combinatorial semiring operations')
print(tCsum(countArrival(a), countArrival(b)))
print(tCprod(countArrival(a), countArrival(b)))
  
print('\n first arrival betweenness, Figure 4')
TQ.VecList(tempBetween(A))

print('--------------------------- FIRST ARRIVAL BETWEENNESS WITH EXCLUSION -----------------------')
print('first arrival betweenness with exclusion, Test 1')
A = TQ.Ianus2Mat('test.ten')['mat']
TQ.VecList(betweenTimeEx(A))
print('----- FAB')
TQ.VecList(tempBetween(A))

print('first arrival betweenness with exclusion, Test 2')
A = TQ.Ianus2Mat('test2.ten')['mat']
TQ.VecList(betweenTimeEx(A))
print('----- FAB')
TQ.VecList(tempBetween(A))

print('first arrival betweenness with exclusion, Test 3')
A = TQ.Ianus2Mat('test3.ten')['mat']
TQ.VecList(betweenTimeEx(A))
print('----- FAB')
TQ.VecList(tempBetween(A))

print('----------------- TEST LPP -------------------------------')
A = TQ.Ianus2Mat('test2LPP.ten')['mat']
start = time.time()
TQ.VecList(betweenTimeEx(A))
end = time.time()
print(' ..... cas .....')
print(end - start)
print('----- FAB')
start = time.time()
TQ.VecList(tempBetween(A))
end = time.time()
print(' ..... cas .....')
print(end - start)